'******************************************************************************
' Version 1.0
' Copyright (c) 2011, Dave Hein
' See end of file for terms of use.
'******************************************************************************
{{
  This object contains methods that are used to access external memory.
}}
con
  SYS_COMMAND = $12340000
  SYS_LOCKNUM = $12340002
  SYS_PARM    = $12340004

  SYS_EXTMEM_READ   = 17
  SYS_EXTMEM_WRITE  = 18
  SYS_EXTMEM_ALLOC  = 19

pub ExtMemRead(ExtMemAddr, HubAddr, NumBytes)
'' This method copies NumBytes of data from external memory starting
'' at the address ExtMemAddr to hub RAM starting at address HubAddr.
'' The return value will indicate the number of bytes transferred.
  result := SystemCall(SYS_EXTMEM_READ, @ExtMemAddr)

pub ExtMemWrite(ExtMemAddr, HubAddr, NumBytes)
'' This method copies NumBytes of data to external memory starting
'' at the address ExtMemAddr from hub RAM starting at address HubAddr.
'' The return value will indicate the number of bytes transferred.
  result := SystemCall(SYS_EXTMEM_WRITE, @ExtMemAddr)

pub ExtMemAlloc(ExtMemAddr, NumBytes)
'' This method allocates NumBytes of data for an external memory
'' starting at ExtMemAddr.  The allocation will fail if there is not
'' enough memory available, or the address space overlaps and existing
'' hub or external memory space or four external memories have already
'' been allocated.  The number of bytes allocated will will be returned
'' on success, and zero will be returned on failure.
  result := SystemCall(SYS_EXTMEM_ALLOC, @ExtMemAddr)

pri SystemCall(command, parm) | locknum
  locknum := word[SYS_LOCKNUM] - 1
  if locknum == -1
    return -1
    
  repeat until not lockset(locknum)
  long[SYS_PARM] := parm
  word[SYS_COMMAND] := command
  repeat while word[SYS_COMMAND]
  result := long[SYS_PARM]
  lockclr(locknum)
{{
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
}}
 