/*******************************************************************************
' Version 0.27
' Copyright (c) 2010, 2011, Dave Hein
' See end of file for terms of use.
'******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#ifdef LINUX
#include <dirent.h>
#include <sys/stat.h>
#include "conion.h"
#else
#include <conio.h>
#include <direct.h>
#endif
#include "interp.h"
#include "rom.h"
#include "spindebug.h"

// Define system I/O addresses and commands
#define SYS_COMMAND    0x12340000
#define SYS_LOCKNUM    0x12340002
#define SYS_PARM       0x12340004
#define SYS_DEBUG      0x12340008

#define SYS_CON_PUTCH     1
#define SYS_CON_GETCH     2
#define SYS_FILE_OPEN     3
#define SYS_FILE_CLOSE    4
#define SYS_FILE_READ     5
#define SYS_FILE_WRITE    6
#define SYS_FILE_OPENDIR  7
#define SYS_FILE_CLOSEDIR 8
#define SYS_FILE_READDIR  9
#define SYS_FILE_SEEK     10
#define SYS_FILE_TELL     11
#define SYS_FILE_REMOVE   12
#define SYS_FILE_CHDIR    13
#define SYS_FILE_GETCWD   14
#define SYS_FILE_MKDIR    15
#define SYS_FILE_GETMOD   16
#define SYS_EXTMEM_READ   17
#define SYS_EXTMEM_WRITE  18
#define SYS_EXTMEM_ALLOC  19

static char rootdir[100];
char *hubram;
char *extram[4];
int32_t extmemsize[4];
uint32_t extmembase[4];
int32_t extmemnum = 0;
char lockstate[8];
char lockalloc[8];

char objname[100][20];
int32_t methodnum[100];
int32_t methodlev = 0;

int32_t printflag = 0;
int32_t symflag = 0;
int32_t pasmspin = 0;
int32_t profile = 0;
int32_t memsize = 64;
int32_t cycleaccurate = 0;
int32_t loopcount = 0;
int32_t proptwo = 0;
int32_t baudrate = 0;
int32_t pin_val = -1;

PasmVarsT PasmVars[8];

void PrintOp(SpinVarsT *spinvars);
void ExecuteOp(SpinVarsT *spinvars);
char *FindChar(char *str, int32_t val);

void usage(void)
{
    printf("SpinSim Version 0.27\n");
    printf("usage: spinsim [options] file\n");
    printf("The options are as follows:\n");
    printf("     -l  List executed instructions\n");
    printf("     -p  Use PASM Spin interpreter\n");
    printf("     -#  Execute # instructions\n");
    printf("     -P  Profile Spin opcode usage\n");
    printf("     -m# Set the hub memory size to # K-bytes\n");
    printf("     -c  Enable cycle-accurate mode for pasm cogs\n");
    //printf("     -t  Enable the Prop 2 mode\n");
    printf("     -b# Enable the serial port and set the baudrate to #\n");
    //printf("     -x# Set the external memory size to # K-bytes\n");
    exit(1);
}

void putchx(int32_t val)
{
    putchar(val);
    fflush(stdout);
}

int32_t getchx(void)
{
    uint8_t val = 0;
    // GCC compiler issues warning for ignored fread return value
    if(fread(&val, 1, 1, stdin))
    	if (val == 10) val = 13;
    return val;
}

char *FindExtMem(uint32_t addr, int32_t num)
{
    int i;
    char *ptr = 0;
    uint32_t addr1 = addr + num - 1;
    uint32_t curraddr, curraddr1;

    //printf("FindExtMem(%d, %d)\n", addr, num);

    for (i = 0; i < extmemnum; i++)
    {
	//printf("i = %d\n", i);
        curraddr = extmembase[i];
        curraddr1 = curraddr + extmemsize[i] - 1;
        if (curraddr <= addr && addr <= curraddr1)
	{
	    //printf("1: %d %d %d\n", addr, curraddr, curraddr1);
	    //printf("2: %d %d %d\n", addr1, curraddr, curraddr1);
            if (curraddr <= addr1 && addr1 <= curraddr1)
	        ptr = extram[i] + addr - extmembase[i];
	    break;
	}
    }

    return ptr;
}

void CheckCommand(void)
{
    int32_t parm;
    int32_t command = WORD(SYS_COMMAND);
    FILE *stream;
    DIR *pdir;
    struct dirent *pdirent;

    if (!command) return;

    parm = LONG(SYS_PARM);

    if (command == SYS_CON_PUTCH)
    {
	if (parm == 13)
	    putchx(10);
	else
	    putchx(parm);
    }
    else if (command == SYS_CON_GETCH)
    {
	if (kbhit())
	    parm = getch();
	else
	    parm = -1;
	LONG(SYS_PARM) = parm;
    }
    else if (command == SYS_FILE_OPEN)
    {
	char *fname;
	char *fmode;

	fname = (char *)&BYTE(LONG(parm));
	fmode = (char *)&BYTE(LONG(parm+4));
	stream = fopen(fname, fmode);
	LONG(SYS_PARM) = (long)stream;
    }
    else if (command == SYS_FILE_CLOSE)
    {
	if (parm) fclose((FILE *)(long)parm);
	LONG(SYS_PARM) = 0;
    }
    else if (command == SYS_FILE_READ)
    {
	int32_t num;
	char *buffer;

         if (!parm) LONG(SYS_PARM) = -1;
	 else
	 {
	    stream = (FILE *)(long)LONG(parm);
	    buffer = (char *)&BYTE(LONG(parm+4));
	    num = LONG(parm+8);
	    LONG(SYS_PARM) = fread(buffer, 1, num, stream);
	}
    }
    else if (command == SYS_FILE_WRITE)
    {
	int32_t num;
	char *buffer;

	if (!parm) LONG(SYS_PARM) = -1;
	else
	{
	    stream = (FILE *)(long)LONG(parm);
	    buffer = (char *)&BYTE(LONG(parm+4));
	    num = LONG(parm+8);
	    LONG(SYS_PARM) = fwrite(buffer, 1, num, stream);
	}
    }
    else if (command == SYS_FILE_OPENDIR)
    {
	char *dname = ".";

	pdir = opendir(dname);
	LONG(SYS_PARM) = (long)pdir;
    }
    else if (command == SYS_FILE_CLOSEDIR)
    {
	if (parm) closedir((DIR *)(long)parm);
	LONG(SYS_PARM) = 0;
    }
    else if (command == SYS_FILE_READDIR)
    {
	int32_t *buffer;
        pdir = (DIR *)(long)LONG(parm);
        buffer = (int32_t *)&BYTE(LONG(parm+4));
	if (!pdir) LONG(SYS_PARM) = 0;
	else
	{
	    pdirent = readdir(pdir);
	    if (pdirent)
	    {
#ifdef LINUX
		FILE *infile;
		int32_t d_size = 0;
		int32_t d_attr = 0;
		int32_t d_type = pdirent->d_type;

		if (d_type & DT_DIR) d_attr |= 0x10;
		if (d_type & S_IXUSR) d_attr |= 0x20;
		if (!(d_type & S_IWUSR)) d_attr |= 0x01;

                if ((infile = fopen(pdirent->d_name, "r")))
		{
		    fseek(infile, 0, SEEK_END);
		    d_size = ftell(infile);
		    fclose(infile);
		}

 	        buffer[0] = d_size;
 	        buffer[1] = d_attr;
#else
	        buffer[0] = pdirent->d_size;
	        buffer[1] = pdirent->d_attr;
#endif
	        strcpy((char *)&buffer[2], pdirent->d_name);
	    }
	    LONG(SYS_PARM) = (long)pdirent;
	}
    }
    else if (command == SYS_FILE_SEEK)
    {
	int32_t offset, whence;
        stream = (FILE *)(long)LONG(parm);
        offset = LONG(parm+4);
        whence = LONG(parm+8);
	LONG(SYS_PARM) = fseek(stream, offset, whence);
    }
    else if (command == SYS_FILE_TELL)
    {
        stream = (FILE *)(long)parm;
	LONG(SYS_PARM) = ftell(stream);
    }
    else if (command == SYS_FILE_REMOVE)
    {
	char *fname = (char *)&BYTE(parm);
	LONG(SYS_PARM) = remove(fname);
    }
    else if (command == SYS_FILE_CHDIR)
    {
	char *path = (char *)&BYTE(parm);
	char fullpath[200];
	char *ptr;
	if (path[0] == '/')
	{
	    strcpy(fullpath, rootdir);
	    strcat(fullpath, path);
	}
	else
	    strcpy(fullpath, path);

	ptr = fullpath;
#ifndef LINUX
	while (*ptr)
	{
	    if (*ptr == '/') *ptr = 0x5c;
	    ptr++;
	}
#endif
	parm = chdir(fullpath);
	LONG(SYS_PARM) = parm;
    }
    else if (command == SYS_FILE_GETCWD)
    {
        char *ptr;
	char *str = (char *)&BYTE(LONG(parm));
	int32_t num = LONG(parm+4);
	ptr = getcwd(str, num);
	LONG(SYS_PARM) = LONG(parm);
    }
    else if (command == SYS_FILE_MKDIR)
    {
	char *fname = (char *)&BYTE(parm);
#ifdef LINUX
	LONG(SYS_PARM) = mkdir(fname, S_IRWXU | S_IRWXG | S_IRWXO);
#else
	LONG(SYS_PARM) = mkdir(fname);
#endif
    }
    else if (command == SYS_FILE_GETMOD)
    {
	char *fname = (char *)&BYTE(parm);
	int32_t attrib = -1;

	pdir = opendir(".");
	while (pdir)
	{
	    pdirent = readdir(pdir);
	    if (!pdirent) break;
	    if (strcmp(pdirent->d_name, fname) == 0)
	    {
#ifdef LINUX
		int32_t d_type = pdirent->d_type;
		attrib = 0;
		if (d_type & DT_DIR) attrib |= 0x10;
		if (d_type & S_IXUSR) attrib |= 0x20;
		if (!(d_type & S_IWUSR)) attrib |= 0x01;
#else
		attrib = pdirent->d_attr;
#endif
		break;
	    }
	}
	if (pdir) closedir(pdir);
	LONG(SYS_PARM) = attrib;
    }
    else if (command == SYS_EXTMEM_READ)
    {
	uint32_t extaddr = (uint32_t)LONG(parm);
	char *hubaddr = (char *)&BYTE(LONG(parm+4));
	int32_t num = LONG(parm+8);
	char *extmemptr = FindExtMem(extaddr, num);
	if (extmemptr)
	{
	    memcpy(hubaddr, extmemptr, num);
	    LONG(SYS_PARM) = num;
	}
	else
	{
	    LONG(SYS_PARM) = 0;
	}
    }
    else if (command == SYS_EXTMEM_WRITE)
    {
	uint32_t extaddr = (int32_t)LONG(parm);
	char *hubaddr = (char *)&BYTE(LONG(parm+4));
	int32_t num = LONG(parm+8);
	char *extmemptr = FindExtMem(extaddr, num);
	if (extmemptr)
	{
	    memcpy(extmemptr, hubaddr, num);
	    LONG(SYS_PARM) = num;
	}
	else
	{
	    LONG(SYS_PARM) = 0;
	}
    }
    else if (command == SYS_EXTMEM_ALLOC)
    {
	uint32_t extaddr = (int32_t)LONG(parm);
	int32_t num = LONG(parm+4);
	uint32_t extaddr1 = extaddr + num - 1;

	if (num <= 0 || extmemnum >= 4) num = 0;
	else
	{
	    int i;
	    uint32_t curraddr, curraddr1;
	    for (i = 0; i < extmemnum; i++)
	    {
	        curraddr = extmembase[i];
	        curraddr1 = curraddr + extmemsize[i] - 1;
	        if (curraddr <= extaddr && extaddr <= curraddr1) break;
	        if (curraddr <= extaddr1 && extaddr1 <= curraddr1) break;
	        if (extaddr <= curraddr && curraddr <= extaddr1) break;
	        if (extaddr <= curraddr1 && curraddr1 <= extaddr1) break;
	    }
	    if (i != extmemnum) num = 0;
	    else
	    {
	        extram[extmemnum] = malloc(num);
	        if (extram[extmemnum] == 0) num = 0;
		else
	        {
		    extmemsize[extmemnum] = num;
		    extmembase[extmemnum] = extaddr;
		    extmemnum++;
	        }
	    }
	}
	LONG(SYS_PARM) = num;
    }
    WORD(SYS_COMMAND) = 0;
}

void CheckSerialIn(void)
{
    static int state = 0;
    static int count;
    static int val;

    if (state == 0)
    {
	if (kbhit())
	{
	    val = getch();
	    val |= 0x300;
	    count = 20000000/baudrate;
	    pin_val &= 0x7fffffff;
	    state = 1;
	}
    }
    else if (--count <= 0)
    {
	if (++state > 11)
	    state = 0;
	else
	{
	    pin_val = (pin_val & 0x7fffffff) | ((val & 1) << 31);
	    val >>= 1;
	    count = 20000000/baudrate;
	}
    }
}

void CheckSerialOut(void)
{
    int txbit = 0;
    static int val;
    static int state = -2;
    static int count;

    txbit = (pin_val >> 30) & 1;

    if (state == -2)
    {
	if (txbit)
	{
	    state = -1;
	    //printf("Start Serial\n");
	}
    }
    else if (state == -1)
    {
	if (!txbit)
	{
	    val = 0;
	    state = 0;
	    count = 20000000*3/(baudrate*2);
	}
    }
    else
    {
	if (--count <= 0)
	{
	    if (state > 7)
	    {
		state = -1;
		if (val == 13)
		    putchx(10);
		else
		    putchx(val);
	    }
	    else
	    {
		//printf("%d", txbit);
	        val |= txbit << state;
	        count = 20000000/baudrate;
		state++;
	    }
	}
    }
}

void PrintStack(SpinVarsT *spinvars)
{
    int32_t dcurr = spinvars->dcurr;
    printf("PrintStack: %4.4x %8.8x %8.8x %8.8x\n",
        dcurr, LONG(dcurr-4), LONG(dcurr-8), LONG(dcurr-12));
}

char *bootfile;

void reboot(void)
{
    int32_t i;
    int32_t dbase;
    char *ptr;
    FILE *infile;

    memset(hubram, 0, 32768);
    memset(lockstate, 0, 8);
    memset(lockalloc, 0, 8);

    infile = fopen(bootfile, "rb");

    if (infile == 0)
    {
	printf("Could not open %s\n", bootfile);
	exit(1);
    }

    i = fread(hubram, 1, 32768, infile);
    fclose(infile);

    // Copy in the ROM contents
    memcpy(hubram + 32768, romdata, 32768);

    dbase = WORD(10);
    LONG(dbase-8) = 0xfff9ffff;
    LONG(dbase-4) = 0xfff9ffff;
    LONG(dbase) = 0;

    WORD(SYS_COMMAND) = 0;
    WORD(SYS_LOCKNUM) = 1;
    lockalloc[0] = 1;

    for (i = 0; i < 8; i++) PasmVars[i].state = 0;

    if (pasmspin)
        StartPasmCog(&PasmVars[0], 0x0004, 0xf004, 0);
    else
        StartCog((SpinVarsT *)&PasmVars[0].mem[0x1e0], 4, 0);

    strcpy(objname[0], "xxx");
    strcpy(objname[1], bootfile);
    ptr = FindChar(objname[1], '.');
    if (*ptr)
    {
        *ptr = 0;
	if (symflag && strcmp(ptr + 1, "bin") == 0)
	  symflag = 2;
    }
    if (symflag == 2)
        strcat(objname[1], ".spn");
    else
        strcat(objname[1], ".spin");
    methodnum[0] = 1;
    methodnum[1] = 1;
    methodlev = 1;

    LONG(SYS_DEBUG) = printflag;
}

int32_t main(int32_t argc, char **argv)
{
    char *ptr;
    char *fname = 0;
    int32_t i;
    int32_t maxloops = -1;
    int32_t runflag;
    SpinVarsT *spinvars;
    int32_t state;

    ptr = getcwd(rootdir, 100);

    for (i = 1; i < argc; i++)
    {
	if (strcmp(argv[i], "-l") == 0)
	    printflag = 255;
	else if (strcmp(argv[i], "-c") == 0)
	    cycleaccurate = 1;
	else if (strcmp(argv[i], "-t") == 0)
            proptwo = 1;
	else if (strcmp(argv[i], "-p") == 0)
	{
	    pasmspin = 1;
	    cycleaccurate = 1;
	}
	else if (strcmp(argv[i], "-s") == 0)
	    symflag = 1;
	else if (strcmp(argv[i], "-P") == 0)
	    profile = 1;
	else if (strncmp(argv[i], "-m", 2) == 0)
	{
	    sscanf(&argv[i][2], "%d", &memsize);
	}
	else if (strncmp(argv[i], "-b", 2) == 0)
	{
	    pasmspin = 1;
	    cycleaccurate = 1;
	    sscanf(&argv[i][2], "%d", &baudrate);
	}
#if 0
	else if (strncmp(argv[i], "-x", 2) == 0)
	{
	    sscanf(&argv[i][2], "%d", &extmemsize);
	}
#endif
	else if (argv[i][0] == '-' && argv[i][1] >= '0' && argv[i][1] <= '9')
	    sscanf(argv[i] + 1, "%d", &maxloops);
	else if (argv[i][0] == '-')
	    usage();
	else if (!fname)
	    fname = argv[i];
	else
	    usage();
    }

    // Check the hub memory size and allocate it
    if (memsize < 32)
    {
	printf("Specified memory size is too small\n");
	exit(1);
    }
    if (memsize < 64) memsize = 64;
    memsize <<= 10; // Multiply it by 1024
    hubram = malloc(memsize + 16 + 3);
    if (!hubram)
    {
	printf("Specified memory size is too large\n");
	exit(1);
    }
    // Make sure it's long aligned
    hubram =  (char *)(((int32_t)hubram) & 0xfffffffc);

#if 0
    // Check the ext memory size and allocate it if non-zero
    if (extmemsize < 0)
    {
	printf("Invalid external memory size\n");
	exit(1);
    }
    else if (extmemsize)
    {
        extmemsize <<= 10; // Multiply it by 1024
        extram = malloc(extmemsize);
        if (!extram)
        {
	    printf("Specified external memory size is too large\n");
	    exit(1);
        }
    }
#endif

    if (profile) ResetStats();

    bootfile = fname;

    if (!fname) usage();

    reboot();

    runflag = 1;
    while (runflag && (maxloops < 0 || loopcount < maxloops))
    {
	runflag = 0;
	for (i = 0; i < 8; i++)
	{
	    state = PasmVars[i].state;
	    if (state & 4)
	    {
	        if ((LONG(SYS_DEBUG) & (1 << i)) && state == 5)
		{
		    printf("Cog %d:  ", i);
		    DebugPasmInstruction(&PasmVars[i]);
		}
	        ExecutePasmInstruction(&PasmVars[i]);
		runflag = 1;
	    }
	    else if (state)
	    {
		spinvars = (SpinVarsT *)&PasmVars[i].mem[0x1e0];
	        if ((LONG(SYS_DEBUG) & (1 << i)) && state == 1)
		{
		    int32_t dcurr = spinvars->dcurr;
		    printf("Cog %d: %4.4x %8.8x - ", i, dcurr, LONG(dcurr - 4));
		    PrintOp(spinvars);
		}
		if (profile) CountOp(spinvars);
	        ExecuteOp(spinvars);
		runflag = 1;
	    }
            CheckCommand();
	}
	if (baudrate)
	{
	    CheckSerialOut();
	    CheckSerialIn();
	}
	loopcount++;
    }
    if (profile) PrintStats();
    return 0;
}
/*
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
*/
